@@ -1,20 +1,11 @@
 #!/bin/sh
 #
-# autogen.sh glue for CMU Cyrus IMAP
-# $Id: autogen.sh,v 1.6 2004/11/08 15:07:22 steinm Exp $
+# autogen.sh glue for pxlib
+# $Id: autogen.sh,v 1.8 2005/08/22 15:05:45 steinm Exp $
 #
 # Requires: automake, autoconf, dpkg-dev
 set -e
 
-# Refresh GNU autotools toolchain.
-for i in config.guess config.sub missing install-sh mkinstalldirs ; do
-	test -r /usr/share/automake/${i} && {
-		rm -f ${i}
-		cp /usr/share/automake/${i} .
-	}
-	chmod 755 ${i}
-done
-
 intltoolize --force --copy
 aclocal
 autoheader
