@@ -119,7 +119,9 @@ pxhead_t *get_px_head(pxdoc_t *pxdoc, pxstream_t *pxs)
 	if(((pxh->px_filetype == pxfFileTypIndexDB) ||
 		  (pxh->px_filetype == pxfFileTypNonIndexDB) ||
 		  (pxh->px_filetype == pxfFileTypNonIncSecIndex) ||
-		  (pxh->px_filetype == pxfFileTypIncSecIndex)) &&
+		  (pxh->px_filetype == pxfFileTypIncSecIndex) ||
+          (pxh->px_filetype == pxfFileTypNonIncSecIndexG) ||
+		  (pxh->px_filetype == pxfFileTypIncSecIndexG)) &&
 		  (pxh->px_fileversion >= 40)) {
 		if((ret = pxdoc->read(pxdoc, pxs, sizeof(TPxDataHeader), &pxdatahead)) < 0) {
 			pxdoc->free(pxdoc, pxh);
@@ -174,8 +176,10 @@ pxhead_t *get_px_head(pxdoc_t *pxdoc, pxstream_t *pxs)
 		return NULL;
 	}
 
-	/* skip the tfieldNamePtrArray, not present in index files */
-	if(pxhead.fileType == 0 || pxhead.fileType == 2) {
+	/* skip the tfieldNamePtrArray, not present in primary index files */
+	if(pxhead.fileType == 0 || pxhead.fileType == 2 ||
+	   pxhead.fileType == 3 || pxhead.fileType == 5 ||
+	   pxhead.fileType == 6 || pxhead.fileType == 8) {
 		for(i=0; i<pxh->px_numfields; i++) {
 			if((ret = pxdoc->read(pxdoc, pxs, sizeof(int), dummy)) < 0) {
 				pxdoc->free(pxdoc, pxh->px_fields);
@@ -245,7 +249,9 @@ int put_px_head(pxdoc_t *pxdoc, pxhead_t *pxh, pxstream_t *pxs) {
 	isindex = !((pxh->px_filetype == pxfFileTypIndexDB) ||
 		        (pxh->px_filetype == pxfFileTypNonIndexDB) ||
 		        (pxh->px_filetype == pxfFileTypNonIncSecIndex) ||
-		        (pxh->px_filetype == pxfFileTypIncSecIndex));
+		        (pxh->px_filetype == pxfFileTypIncSecIndex) ||
+                (pxh->px_filetype == pxfFileTypNonIncSecIndexG) ||
+		        (pxh->px_filetype == pxfFileTypIncSecIndexG));
 
 	put_short_le((char *)&pxhead.recordSize, pxh->px_recordsize);
 	put_short_le((char *)&pxhead.headerSize, pxh->px_headersize);
@@ -306,6 +312,7 @@ int put_px_head(pxdoc_t *pxdoc, pxhead_t *pxh, pxstream_t *pxs) {
 			break;
 	}
 	pxhead.sortOrder = pxh->px_sortorder;
+	pxhead.refIntegrity = pxh->px_refintegrity;
 	if(!isindex && (pxh->px_fileversion >= 40)) {
 		dataheadoffset = 0x78;
 	} else {
@@ -407,7 +414,10 @@ int put_px_head(pxdoc_t *pxdoc, pxhead_t *pxh, pxstream_t *pxs) {
 	pxf = pxh->px_fields;
 	for(i=0; i<pxh->px_numfields; i++, pxf++) {
 		pxinfo.fType = pxf->px_ftype;
-		pxinfo.fSize = pxf->px_flen;
+		if(pxf->px_ftype == pxfBCD)
+			pxinfo.fSize = pxf->px_fdc;
+		else
+			pxinfo.fSize = pxf->px_flen;
 		if(pxdoc->write(pxdoc, pxs, sizeof(TFldInfoRec), &pxinfo) < 1) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not write field specification."));
 			return -1;
@@ -447,7 +457,7 @@ int put_px_head(pxdoc_t *pxdoc, pxhead_t *pxh, pxstream_t *pxs) {
 	/* write tablename */
 	if(pxh->px_tablename == NULL) {
 		len = 0;
-		px_error(pxdoc, PX_Warning, _("Tablename is empty."));
+//		px_error(pxdoc, PX_Warning, _("Tablename is empty."));
 	} else {
 		len = strlen(pxh->px_tablename);
 		if(pxdoc->write(pxdoc, pxs, len, pxh->px_tablename) < 1) {
@@ -653,12 +663,30 @@ int put_px_datablock(pxdoc_t *pxdoc, pxhead_t *pxh, int after, pxstream_t *pxs)
 /* }}} */
 
 /* px_add_data_to_block() {{{
- * stores a record into a data block. datablocknr is the logical number
+ * stores a record into a data block. datablocknr is the physical number
  * of the block (the first block has number 1).
+ * recnr is the number of the record within the block. The first record
+ * in a block has number 0.
+ * update is set to 1 if an existing record is updated otherwise it will
+ * be set to 0.
+ * The function returns the number of records in the modified block. This
+ * is either n+1 or n, depending on whether a new record was added or
+ * an exiting record was updated.
+ * -1 is returned in case of an error.
  */
-int px_add_data_to_block(pxdoc_t *pxdoc, pxhead_t *pxh, int datablocknr, char *data, pxstream_t *pxs) {
+int px_add_data_to_block(pxdoc_t *pxdoc, pxhead_t *pxh, int datablocknr, int recnr, char *data, pxstream_t *pxs, int *update) {
 	TDataBlock datablockhead;
-	int ret, n;
+	int ret, n, pos;
+
+	int recsperdatablock = (pxdoc->px_head->px_maxtablesize*0x400-sizeof(TDataBlock)) / pxdoc->px_head->px_recordsize;
+	if(recnr < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not write a record into a block, because the record position is less than 0."));
+		return -1;
+	}
+	if(recnr >= recsperdatablock) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not write a record into a block, because the record position is greater than or equal the maximum number of records per block."));
+		return -1;
+	}
 
 	/* Get header of the data block */
 	if((ret = get_datablock_head(pxdoc, pxs, datablocknr, &datablockhead)) < 0) {
@@ -666,25 +694,49 @@ int px_add_data_to_block(pxdoc_t *pxdoc, pxhead_t *pxh, int datablocknr, char *d
 		return -1;
 	}
 
-	n = get_short_le((char *) &datablockhead.addDataSize)/pxh->px_recordsize+1;
+	/* Use get_short_le_s() instead of get_short_le() because the value
+	 * is negative if the block has no records.
+	 */
+	n = get_short_le_s((char *) &datablockhead.addDataSize)/pxh->px_recordsize + 1;
 //	fprintf(stderr, "Hexdump des alten datablock headers: ");
 //	hex_dump(stderr, &datablockhead, sizeof(TDataBlock));
 //	fprintf(stderr, "\n");
-//	fprintf(stderr, "Gre des Datenblocks: %d\n", get_short_le((char *) &datablockhead.addDataSize));
+//	fprintf(stderr, "Gre des Datenblocks: %d\n", get_short_le_s((char *) &datablockhead.addDataSize));
 //	fprintf(stderr, "Datablock %d has %d records\n", datablocknr, n);
+//	fprintf(stderr, "Adding new record at postion %d in block\n", recnr);
 
 	/* Update size of data block and write it back */
 //	fprintf(stderr, "Hexdump des neuen datablock headers: ");
 //	hex_dump(stderr, &datablockhead, sizeof(TDataBlock));
 //	fprintf(stderr, "\n");
-	put_short_le((char *)&datablockhead.addDataSize, n*pxh->px_recordsize);
-	if(put_datablock_head(pxdoc, pxs, datablocknr, &datablockhead) < 0) {
-		px_error(pxdoc, PX_RuntimeError, _("Could not write updated data block header."));
-		return -1;
+
+	/* Check if the requested record number is not right behind the last
+	 * record in the block. Fix the record number if needed
+	 */
+	if(recnr > n)
+		recnr = n;
+
+	/* Check if record number within the block is larger then the current
+	 * number of records-1 in the block. If yes, we need to increment the
+	 * number of records, otherwise we simply overwrite an existing record.
+	 */
+	if(recnr == n) {
+		n++;
+//		fprintf(stderr, "Set new number of records in block to %d\n", n);
+		put_short_le((char *)&datablockhead.addDataSize, (n-1)*pxh->px_recordsize);
+		if(put_datablock_head(pxdoc, pxs, datablocknr, &datablockhead) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not write updated data block header."));
+			return -1;
+		}
+		*update = 0;
+		pos = n-1;
+	} else {
+		pos = recnr;
+		*update = 1;
 	}
 
 	/* Goto start of record data */
-	if((ret = pxdoc->seek(pxdoc, pxs, n*pxh->px_recordsize, SEEK_CUR)) < 0) {
+	if((ret = pxdoc->seek(pxdoc, pxs, pxh->px_headersize+(datablocknr-1)*pxh->px_maxtablesize*0x400+sizeof(TDataBlock)+pos*pxh->px_recordsize, SEEK_SET)) < 0) {
 		px_error(pxdoc, PX_RuntimeError, _("Could not fseek to start of new record."));
 		return -1;
 	}
@@ -695,7 +747,232 @@ int px_add_data_to_block(pxdoc_t *pxdoc, pxhead_t *pxh, int datablocknr, char *d
 		return -1;
 	}
 	
-	return n;
+	return pos;
+}
+/* }}} */
+
+/* px_delete_data_from_block() {{{
+ * deletes a record from a data block. datablocknr is the physical number
+ * of the block (the first block has number 1).
+ * recnr is the number of the record within the block. The first record
+ * in a block has number 0.
+ * The block is reorganized to make sure all records are stored at
+ * the beginning of the block. If record n of m records is deleted,
+ * then the records n+1 up to m will be copied to n up to m-1.
+ * The function returns the remaining number of records in the block,
+ * which should be 1 less than before.
+ * A value less than 0 is returned in case of an error.
+ */
+int px_delete_data_from_block(pxdoc_t *pxdoc, pxhead_t *pxh, int datablocknr, int recnr, pxstream_t *pxs) {
+	TDataBlock datablockhead;
+	int ret, n, i;
+	char *data;
+
+	int recsperdatablock = (pxdoc->px_head->px_maxtablesize*0x400-sizeof(TDataBlock)) / pxdoc->px_head->px_recordsize;
+	if(recnr < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not write a record into a block, because the record position is less than 0."));
+		return -1;
+	}
+	if(recnr >= recsperdatablock) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not write a record into a block, because the record position is greater than or equal the maximum number of records per block."));
+		return -2;
+	}
+
+	/* Get header of the data block */
+	if((ret = get_datablock_head(pxdoc, pxs, datablocknr, &datablockhead)) < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not read data block header."));
+		return -3;
+	}
+
+	n = get_short_le_s((char *) &datablockhead.addDataSize)/pxh->px_recordsize;
+
+	/* Check if record number within the block is less or equal the current
+	 * number of records-1 in the block. If yes, we need to decrement the
+	 * number of records and pack the data block, otherwise we though an error.
+	 */
+	if(recnr <= n) {
+		n--;
+	} else {
+		px_error(pxdoc, PX_RuntimeError, _("The record number of the record to be deleted is beyond the number of records in the data block: %d:%d < %d."), datablocknr, recnr, n);
+		return -4;
+	}
+
+	/* Write header of data block */
+	put_short_le((char *)&datablockhead.addDataSize, n*pxh->px_recordsize);
+	if(put_datablock_head(pxdoc, pxs, datablocknr, &datablockhead) < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not write updated data block header."));
+		return -5;
+	}
+
+	/* If the last record in the block was deleted, than there is no need
+	 * to repack the block
+	 */
+	if(recnr == n+1) {
+		return n+1;
+	}
+
+	/* Jump to start of delete record */
+	if((ret = pxdoc->seek(pxdoc, pxs, recnr*pxh->px_recordsize, SEEK_CUR)) < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not fseek to start of delete record."));
+		return -6;
+	}
+
+	if((data = (char *) pxdoc->malloc(pxdoc, pxh->px_recordsize, _("Allocate memory for temporary record."))) == NULL) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not allocate memory for temporary record."));
+		return -7;
+	}
+
+	for(i=recnr; i<=n; i++) {
+		/* Goto start of next record */
+		if((ret = pxdoc->seek(pxdoc, pxs, pxh->px_recordsize, SEEK_CUR)) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not fseek to start of next record."));
+			pxdoc->free(pxdoc, data);
+			return -8;
+		}
+
+		/* Read data of next record */
+		if((ret = pxdoc->read(pxdoc, pxs, pxh->px_recordsize, data)) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not read next record."));
+			pxdoc->free(pxdoc, data);
+			return -9;
+		}
+
+		/* Go back to deleted record */
+		if((ret = pxdoc->seek(pxdoc, pxs, -2*pxh->px_recordsize, SEEK_CUR)) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not fseek to start of previous record."));
+			pxdoc->free(pxdoc, data);
+			return -10;
+		}
+
+		/* Write the record data */
+		if(pxdoc->write(pxdoc, pxs, pxh->px_recordsize, data) < 1) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not write temporary record."));
+			pxdoc->free(pxdoc, data);
+			return -11;
+		}
+	}
+	pxdoc->free(pxdoc, data);
+
+	return n+1;
+}
+/* }}} */
+
+/* px_delete_blob_data() {{{
+ * deletes a blob from the blob file
+ */
+int px_delete_blob_data(pxblob_t *pxblob, int hsize, int size, int bloboffset, int index) {
+	pxdoc_t *pxdoc;
+	int ret, blocknumber;
+	unsigned char head[12];
+
+	pxdoc = pxblob->pxdoc;
+
+	if((ret = pxblob->seek(pxblob, pxblob->mb_stream, bloboffset, SEEK_SET)) < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
+		return -1;
+	}
+
+	/* Just read the first 3 Bytes because they are common for all block */
+	if((ret = pxblob->read(pxblob, pxblob->mb_stream, 3, head)) < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not read head of blob data."));
+		return -1;
+	}
+
+	if(head[0] == 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Trying to read blob data from 'header' block."));
+		return -1;
+	} else if(head[0] == 4) {
+		px_error(pxdoc, PX_RuntimeError, _("Trying to read blob data from a 'free' block."));
+		return -1;
+	}
+
+	if(head[0] == 2) { /* Deleting blob from a block type 2 */
+		int i, numblocks;
+
+		if(index != 0xff) {
+			px_error(pxdoc, PX_RuntimeError, _("Offset points to a single blob block but index field is not 0xff."));
+			return -1;
+		}
+		/* Read the remaining 6/14 bytes from the header */
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, hsize-3, head)) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not read remaining head of single data block."));
+			return -1;
+		}
+		if(size != get_long_le((char*) &head[0])) {
+			px_error(pxdoc, PX_RuntimeError, _("Blob does not have expected size (%d != %d)."), size, get_long_le((char *) &head[0]));
+			return -1;
+		}
+
+		head[0] = 4;
+		blocknumber = bloboffset >> 12;
+		numblocks = ((size-1) >> 12) + 1;
+		for(i=0; i<numblocks; i++) {
+			if((ret = pxblob->seek(pxblob, pxblob->mb_stream, (blocknumber+i)*0x1000, SEEK_SET)) < 0) {
+				px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
+				return -1;
+			}
+			if((ret = pxblob->write(pxblob, pxblob->mb_stream, 1, &head)) < 0) {
+				px_error(pxdoc, PX_RuntimeError, _("Could not write blob type."));
+				return -1;
+			}
+			pxblob->blocklist[blocknumber+i].type = 4;
+		}
+	} else if(head[0] == 3) { /* Deleting blob from a block type 3 */
+		unsigned char *tmpblock1, *tmpblock2;
+		TMbBlockHeader3Table *tableptr;
+		int offset, i;
+
+		if(NULL == (tmpblock1 = pxdoc->malloc(pxdoc, 2*4096, _("Allocate memory for temporary block from blob file.")))) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not allocate memory for temporary block from blob file."));
+			return -1;
+		}
+
+		tmpblock2 = tmpblock1+4096;
+		memcpy(tmpblock1, head, 3);
+
+		/* Read the the rest of the block */
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, 4096-3, &tmpblock1[3])) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not read remaining data of suballocated block."));
+			return -1;
+		}
+		memcpy(tmpblock2, tmpblock1, 4096);
+
+		/* Goto the blob pointer with the passed index */
+		memcpy(head, tmpblock1+12+index*5, 5);
+		memset(tmpblock1+12+index*5, 0, 5);
+
+		if(size != ((int)head[1]-1)*16+head[4]) {
+			px_error(pxdoc, PX_RuntimeError, _("Blob does not have expected size (%d != %d)."), size, ((int)head[1]-1)*16+head[4]);
+			return -1;
+		}
+
+		tableptr = (TMbBlockHeader3Table *) (tmpblock1+12);
+		offset = 0x15;
+		for(i=63; i>=0; i--) {
+			if(tableptr[i].offset != 0) {
+				memcpy(tmpblock1+offset*16, tmpblock2+tableptr[i].offset*16, tableptr[i].length*16);
+				tableptr[i].offset = offset;
+				offset += tableptr[i].length;
+			}
+		}
+
+		/* Goto the start of the blob */
+		if((ret = pxblob->seek(pxblob, pxblob->mb_stream, bloboffset, SEEK_SET)) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
+			return -1;
+		}
+		if((ret = pxblob->write(pxblob, pxblob->mb_stream, 4096, tmpblock1)) < 0) {
+			px_error(pxdoc, PX_RuntimeError, _("Could not write data of suballocated block."));
+			return -1;
+		}
+
+		blocknumber = bloboffset >> 12;
+		pxblob->blocklist[blocknumber].allocspace -= head[1];
+		pxblob->blocklist[blocknumber].numblobs--;
+
+		pxdoc->free(pxdoc, tmpblock1);
+	}
+	return 0;
 }
 /* }}} */
 
@@ -718,8 +995,10 @@ mbhead_t *get_mb_head(pxblob_t *pxblob, pxstream_t *pxs) {
 		px_error(pxdoc, PX_RuntimeError, _("Could not allocate memory for document header."));
 		return NULL;
 	}
-	if(pxblob->seek(pxblob, pxs, 0, SEEK_SET) < 0)
+	if(pxblob->seek(pxblob, pxs, 0, SEEK_SET) < 0) {
+		px_error(pxdoc, PX_RuntimeError, _("Could not go to start of blob file."));
 		return NULL;
+	}
 	if((ret = pxblob->read(pxblob, pxs, sizeof(TMbHeader), &mbhead)) < 0) {
 		px_error(pxdoc, PX_RuntimeError, _("Could not read header from paradox file."));
 		pxdoc->free(pxdoc, mbh);
